#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <string.h>
#include <pthread.h>  
// (there are more library specific includes later)

typedef enum {NOskel, Pskel, Dskel, Fskel, Bskel, Hskel} skeltag_t; 
// the skeleton tag: no skel, Pipeline, Deal, Farm, Butterfly or HaloSwap

typedef enum {SPLOCAL, SPGLOBAL} spread_t;
// The data spread (local or global)

typedef enum {BUF, STRM} Dmode_t;
// The data mode (buffer or stream)

typedef enum {IMPL, EXPL, DEV} Imode_t; 
// The interaction mode (implicit, explicit or devolved)
// Devolved implies that there is a skeleton persistently nested,
// and that selection of its interaction mode is devolved to the nested skeleton



// The types involved in exchanging and collating the call tree for a hierarchy of 
// calls rooted by a BUF mode call

// Information gathered by a process about its call stack. These are all-all exchanged
// before being used to build the local copy of the overall call tree.
// This is just an int * because these are just blocks of ints. However, the internal
// structure is as follows
//
// | #calls |<---- one call----->|<----- ....more calls ....----->| my id in leaf activity mycomm |
//
// where each call frame is structured as follows
// 
// | skel tag | myid in skel comm | # activities | my act number | my act sz |
//
// so there are 5 ints in a frame and (5*#calls)+2 ints in the whole thing

typedef int * mystackinfo_t;


// A type for the nodes in the locally constructed copies of the global call tree
typedef struct callnode {
  skeltag_t    skel;    // which skeleton (or NOskel for leaf activity)
  MPI_Comm     comm;    // communicator for the call or leaf activity (NULL except on my path)
  int          Nacts;   // number of sub activities (0 for a leaf activity)
  int          level;   // depth in the call tree, counting from 0 for the root BUFF mode call
  int         *actsz;   // for a skel this is a dynamically allocated array of sub activity sizes 
                        // for a leaf activity it is a one element array containing the activity size
                        // duplicated from the in the node above, but useful for indicating the length
                        // of members[0], as described below for the leaf special case
  int        **members; // dynamically allocated array of ptrs to int arrays. 
                        // For a skel node, members[i] points to info on the members of 
                        // sub activity i so that the indicated array A (for sub-act i) contains (level+1)*(actsz[i])
                        // entries, where A[(level+1)*m + l] contains the myid of member m of subact i at level l
                        // (think of it as a 2D array with 1D indexing).
                        // For a leaf activity node there are no subactivities, so members has only one
                        // entry, members[0] which points to an array containing the id histories of 
                        // all members, as above, just like the entries in the parent skel node, but with
                        // the id *within* this activity tacked on for each member.

  struct callnode    **acts;    // dynamically allocated array of pointers 
                                // to nodes for the activities (null for a leaf activity)
  int Gcount, Tcount, Ecount;   // On the trunk only, counters for next Give/Take/Exchange at this 
                                // level. Unused for call tree root node (info is in the trunk child activity)
                                // because the tree has one more level than counter sets. Initialised to 0.
} callnode_t;



/**************************************************************************************************/

/* structural information about a pipeline */
typedef struct {
  int      ns;        // number of stages
  spread_t *spr;      // array of spreads
  MPI_Datatype *ty;   // array of types
  Imode_t  *imode;    // array of interaction modes
  int      shutdown;
  int      *allcols;  // array of stage ids, as selected by all pipeline members
  int      *itemsz;   // ns+1 sizes of raw stage to stage items

} p_t;

/* structural information about a deal */
typedef struct {
  MPI_Comm tofarmercomm;    // for sending messages to the farmer (for buffers)
  MPI_Comm fromfarmercomm;  // for receiving messages from the farmer (for buffers)

  int      nw;          // number of workers
  spread_t inspr;       // spread of the input data
  spread_t outspr;      // spread of the output data
  int      shutdown; 
  int      *allcols;    // array of stage ids, as selected by all deal members
} d_t;

/* the skel_t structure */
typedef struct {
  // Stuff about the skeleton
  skeltag_t skel;
  MPI_Comm  parentcomm;
  int       parentcommsize;
  int       parentcommrank;
  Dmode_t   dmode;          // data mode          
  MPI_Comm  allcomm1;       // a duplicate of parentcomm, for group2group      
  MPI_Comm  allcomm2;       // another duplicate of parentcomm, for group2group 
  int      *allranks;       // all the ranks in parentcomm, in order  
  
  // Stuff about the activity created for this process
  MPI_Comm  mycomm;
  int       mycommsize;
  int       myrank;
  int       mycol;          // my colour, or appropriate default value if implicit

  // We need a comm containing only the root of the stage
  MPI_Comm  stagerootonlycomm;

  //Stuff about the top level buffers
  void *in;
  int  inmul;
  int *inlens;
  MPI_Datatype inty;
  int initemsz;
  spread_t inbufspr;
  int inbufcount; // count the number of access to the input buffer (used in Deal skeleton)

  void *out;
  int **outlenP;
  int **outmulP;
  int outbuffsz;
  MPI_Datatype outty;
  int outitemsz;
  spread_t outbufspr;
  int outbufcount; // count the number of access to the output buffer (used in Deal skeleton)

  char     **buff;         /* the real input buffer, active in all input buffer activity processes */
                           /* for SPGLOBAL data or the root only for SPLOCAL data   */
  int      *blens;         /* local lengths of the mult contributions to the buffer */
                           /* or root only lengths for SP_LOCAL data                */
  int      bnext;          /* marker for running through buffer                     */

  char     **poutbuff;     /* the output buffer, active in all output buffer activity processes*/
  int      *poutblens;     /* local lengths of the poutbnext contributions to this buffer */
  int      poutbnext;      /* marker for running through buffer, counts size              */


  callnode_t *mycalltree;    // NULL except for the innermost skeletons of BUFF rooted nests
  int         nestlevel;     // How many levels are we down the innermost BUFF nest?
                             // 1 for "in the BUFF mode call"

  int nbgive;  // to keep track of the number of give and take performed
  int nbtake;  // (stored always at the Top level)

  union {
    p_t  p;  // Pipeline
    d_t  d;  // Deal
  } context;

} skel_t;


// eDM related types
typedef struct {
  void *data;
  int   len;
} eSkel_atom_t;

typedef struct {
  void **data;
  int   *len;
} eSkel_molecule_t;


/*************************************************************************************************/
// Some definitions

#define MAXDEPTH 6
#define MAXOUTMUL 100000      
#define NOSHUTDOWNTAG 1      /* tag for Give-Take to indicate more data to come */
#define ACKTAG  3            /* tag for acknowledgment */
#define ACKTAG2 4            /* other tag for acknowledgment */
#define FARMTAKETAG 2        /* tag with which a worker requests a task of a farmer */
#define SHUTDOWNTAG 0        /* tag for Give-Take to indicate no more data to come */
#define SHUTDOWNLEN -999     /* tag for Give-Take to indicate no more data to come */
#define ENDSHUTDOWN 10       /* tag to tell that the extra shutdown messages for Deal have been sent */
skel_t *FrStack[MAXDEPTH];

#define SENDER   0
#define RECEIVER 1

#define FALSE 0 
#define TRUE  1

int *somewhere;               /* because I need a safe non-null address to return from SPLOCAL takes */
int mpiintsize;
int zero = 0;
int Top;

// Generic functions  (defined in eSkel_generic.c)
void SkelLibInit (void);
int myrank (void);
int mycommsize (void);
MPI_Comm mycomm (void);

// Generic interaction functions (defined in eSkel_generic.c)
void Give (void *out[], int len[]);
void **Take (int **len);
void *Exchange (int *inlens, void *out, int outlen); // Still old style

// Pipeline functions (defined in pipeline*.c)
void Pipeline (int ns, Imode_t imode[], eSkel_molecule_t * (*stages[])(eSkel_molecule_t *), int col,
               Dmode_t dmode, spread_t spr[], MPI_Datatype ty[], void *in,
               int inlen, int inmul, void *out, int outlen, int *outmul,
               int outbuffsz, MPI_Comm comm);

void Pipelinev (int ns, Imode_t imode[], eSkel_molecule_t * (*stages[])(eSkel_molecule_t *), int col, 
                Dmode_t dmode, spread_t spr[], MPI_Datatype ty[], void *in, int *inlens, int inmul,
                void *out, int *outlens, int *outmul, MPI_Comm comm);

void PipeGive (void *out[], int len[]);
void **PipeTake (int **items);

// Deal functions (defined in deal*.c)
void Deal (int nw, Imode_t imode, eSkel_molecule_t *worker (eSkel_molecule_t *),
	   int col, Dmode_t dmode,
	   void *in, int inlen, int inmul, spread_t inspr, MPI_Datatype inty,
	   void *out, int outlen, int *outmul, spread_t outspr, MPI_Datatype outty,
	   int outbuffsz, MPI_Comm comm);

void DealGive (void *out[], int len[]);
void **DealTake (int **items);


/*************************************************************************************************/
// Library specific includes

#include "eSkel_generic.c"
#include "eSkel_utils.c"
#include "eSkel_calltree.c"

#include "pipelinesupport.c"
#include "pipelinev.c"
#include "pipeline.c"

#include "dealsupport.c"
#include "deal.c"

